#include <J3ML/LinearAlgebra/EulerAngle.hpp>
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <cmath>
#include <algorithm>


namespace J3ML::LinearAlgebra {

    EulerAngleXYZ::EulerAngleXYZ(float roll, float pitch, float yaw) {
        this->roll = roll;
        this->pitch = pitch;
        this->yaw = yaw;
    }

    EulerAngleXYZ::EulerAngleXYZ(const AxisAngle& rhs) {
        *this = EulerAngleXYZ(Quaternion(rhs));
    }

    EulerAngleXYZ::EulerAngleXYZ(const Quaternion& q) {
        float sy = 2 * q.x * q.z + 2 * q.y * q.w;
        bool gimbal_lock = std::abs(sy) > 0.99999f;

        if (!gimbal_lock)
            roll = Math::Degrees(std::atan2(-(2 * q.y * q.z - 2 * q.x * q.w),2 * q.w * q.w + 2 * q.z * q.z - 1));
        else
            roll = Math::Degrees(std::atan2(2 * q.y * q.z + 2 * q.x * q.w,2 * q.w * q.w + 2 * q.y * q.y - 1));

        pitch = Math::Degrees(std::asin(sy));

        if (!gimbal_lock)
            yaw = Math::Degrees(std::atan2(-(2 * q.x * q.y - 2 * q.z * q.w),2 * q.w * q.w + 2 * q.x * q.x - 1));
        else
            yaw = 0;
    }

    EulerAngleXYZ::EulerAngleXYZ(const Matrix3x3& rhs) {
        auto m = rhs.Transposed();
        auto sy = m.At(0, 2);
        auto unlocked = std::abs(sy) < 0.99999f;

        roll = Math::Degrees(unlocked ? std::atan2(-m.At(1, 2), m.At(2, 2)) : std::atan2(m.At(2, 1), m.At(1, 1)));
        pitch = Math::Degrees(std::asin(sy));
        yaw = Math::Degrees(unlocked ? std::atan2(-m.At(0, 1), m.At(0, 0)) : 0);
    }
}